from enum import Enum


class EnergyPatterns(Enum):
    energy_pattern_17 = {'energy_pattern_idx': 'ENERGY_PATTERN_17', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 3 beacons on channel 37, 38 and 39\\r\\n'}
    energy_pattern_18 = {'energy_pattern_idx': 'ENERGY_PATTERN_18', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_39_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch39 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_24 = {'energy_pattern_idx': 'ENERGY_PATTERN_24', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_37_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch37 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_25 = {'energy_pattern_idx': 'ENERGY_PATTERN_25', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_38_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch38 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_26 = {'energy_pattern_idx': 'ENERGY_PATTERN_26', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['2454'], 'sub1g_energy': False, 'desc': 'Energizing on 2454 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_27 = {'energy_pattern_idx': 'ENERGY_PATTERN_27', 'calibration_type': 'CALIBRATION_TEST', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 4 beacons on channels 37,38,39,39 2ms apart\\r\\n'}
    energy_pattern_29 = {'energy_pattern_idx': 'ENERGY_PATTERN_29', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '3', 'energy_channels': ['CHANNEL_38_FREQ', '2454', 'CHANNEL_39_FREQ'], 'sub1g_energy': False, 'desc': 'Evenely energizing on 2426, 2454 and 2480, preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_33 = {'energy_pattern_idx': 'ENERGY_PATTERN_33', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2454'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2454\\r\\n'}
    energy_pattern_34 = {'energy_pattern_idx': 'ENERGY_PATTERN_34', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2427'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2427\\r\\n'}
    energy_pattern_35 = {'energy_pattern_idx': 'ENERGY_PATTERN_35', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2481'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2481\\r\\n'}
    energy_pattern_37 = {'energy_pattern_idx': 'ENERGY_PATTERN_37', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '1', 'energy_channels': ['2450'], 'sub1g_energy': False, 'desc': 'Energizing in 2450MHz. Beacons (0dBm): 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_39 = {'energy_pattern_idx': 'ENERGY_PATTERN_39', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_27_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing in 2460MHz\\r\\n'}
    energy_pattern_40 = {'energy_pattern_idx': 'ENERGY_PATTERN_40', 'calibration_type': 'CALIBRATION_EXIT_TM', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 4 beacons on channels 37,38,22,22 for 450us beacon duration\\r\\n'}
    energy_pattern_41 = {'energy_pattern_idx': 'ENERGY_PATTERN_41', 'calibration_type': 'MINI_RX', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons based on mini rx setting\\r\\n'}
    energy_pattern_42 = {'energy_pattern_idx': 'ENERGY_PATTERN_42', 'calibration_type': 'ENTER_TEST_MODE_01', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2406\\r\\n'}
    energy_pattern_43 = {'energy_pattern_idx': 'ENERGY_PATTERN_43', 'calibration_type': 'ENTER_TEST_MODE_07', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2418\\r\\n'}
    energy_pattern_44 = {'energy_pattern_idx': 'ENERGY_PATTERN_44', 'calibration_type': 'ENTER_TEST_MODE_27', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2460\\r\\n'}
    energy_pattern_47 = {'energy_pattern_idx': 'ENERGY_PATTERN_47', 'calibration_type': 'EXIT_TEST_MODE_01', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2406\\r\\n'}
    energy_pattern_48 = {'energy_pattern_idx': 'ENERGY_PATTERN_48', 'calibration_type': 'EXIT_TEST_MODE_07', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2418\\r\\n'}
    energy_pattern_49 = {'energy_pattern_idx': 'ENERGY_PATTERN_49', 'calibration_type': 'EXIT_TEST_MODE_27', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2460\\r\\n'}
    energy_pattern_50 = {'energy_pattern_idx': 'ENERGY_PATTERN_50', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 915Mhz, beacon train is generated right after the energy pattern\\r\\n '}
    energy_pattern_52 = {'energy_pattern_idx': 'ENERGY_PATTERN_52', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 915Mhz only\\r\\n'}
    energy_pattern_61 = {'energy_pattern_idx': 'ENERGY_PATTERN_61', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 917.5Mhz. Beacons: Neg12dBm: 37 38 39 0dBm: 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_62 = {'energy_pattern_idx': 'ENERGY_PATTERN_62', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '1', 'energy_channels': ['2450'], 'sub1g_energy': True, 'desc': 'Energizing in 2450MHz and Sub1g. Beacons (0dBm): 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_99 = {'energy_pattern_idx': 'ENERGY_PATTERN_99', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '40', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Sweep over BLE channels with LBT\\r\\n'}
